from typing import TypeVar, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from .avrojson import AvroJsonConverter

TC = TypeVar("TC", bound="DictWrapper")


class DictWrapper:
    __slots__ = ["_inner_dict"]
    _inner_dict: dict

    def __init__(self):
        self._inner_dict = {}

    @classmethod
    def construct(cls: Type[TC], inner_dict: dict) -> TC:
        """
        Construct an object without any validations or type annotation checks.
        You should not be using this under normal circumstances.
        """
        obj = cls.__new__(cls)
        obj._inner_dict = {}
        obj._restore_defaults()
        for key, value in inner_dict.items():
            if value:
                obj._inner_dict[key] = value
        return obj

    @classmethod
    def _get_json_converter(cls) -> "AvroJsonConverter":
        # This attribute will be set by the AvroJsonConverter's init method.
        return cls._json_converter

    @classmethod
    def from_obj(cls: Type[TC], obj: dict, tuples: bool = False) -> TC:
        conv = cls._get_json_converter().with_tuple_union(tuples)
        return conv.from_json_object(obj, cls.RECORD_SCHEMA)

    def to_obj(self, tuples: bool = False) -> dict:
        conv = self._get_json_converter().with_tuple_union(tuples)
        return conv.to_json_object(self, self.RECORD_SCHEMA)

    def to_avro_writable(self, fastavro: bool = False) -> dict:
        return self.to_obj(tuples=fastavro)

    def _restore_defaults(self) -> None:
        pass

    def validate(self) -> bool:
        """
        Checks the current object against its pre-defined schema. This does
        not ensure that an object is completely valid (e.g. we don't check that
        the URNs are formatted corrected and point to valid objects), but only
        checks it against the Avro schema. Returns True if valid.
        """
        conv = self._get_json_converter()
        return conv.validate(self.RECORD_SCHEMA, self)

    def get(self, item, default=None):
        return self._inner_dict.get(item, default)

    def items(self):
        return self._inner_dict.items()

    def keys(self):
        return self._inner_dict.keys()

    def __str__(self):
        return f"{self.__class__.__name__}({self._inner_dict.__str__()})"

    def __repr__(self):
        return f"{self.__class__.__name__}({self._inner_dict.__repr__()})"

    def __eq__(self, other):
        if isinstance(other, DictWrapper):
            return self._inner_dict.__eq__(other._inner_dict)
        return False

    def __ne__(self, other):
        return not self.__eq__(other)



def make_fullname(ns, name):
    return ((ns + '.') if ns else '') + name


def split_fullname(fullname):
    idx = fullname.rfind('.')
    if idx < 0:
        return '', fullname

    return fullname[:idx], fullname[idx + 1:]


def get_shortname(fullname):
    return split_fullname(fullname)[1]



import six

class TabbedWriter(object):
    class Indent(object):
        def __init__(self, writer):
            self.writer = writer

        def __enter__(self):
            self.writer.tab()

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.writer.untab()

    def __init__(self, inner_writer, tab_symbol='    '):
        self.__inner_writer = inner_writer
        self.__tabs = 0
        self.__tab_symbol = tab_symbol
        self.__current_tab = ''

    def write(self, text):
        assert isinstance(text, six.string_types)

        start_pos = 0
        last_pos = text.find('\n')

        while last_pos >= 0:
            self.__inner_writer.write(text[start_pos:last_pos + 1])
            self.__inner_writer.write(self.__current_tab)
            start_pos = last_pos + 1
            last_pos = text.find('\n', start_pos)

        self.__inner_writer.write(text[start_pos:])

    def tab(self):
        self.set_tab(self.__tabs + 1)

    def untab(self):
        self.set_tab(self.__tabs - 1)

    def set_tab(self, tabs):
        self.__tabs = max(0, tabs)
        self.__current_tab = self.__tab_symbol * self.__tabs

    def indent(self):
        return TabbedWriter.Indent(self)



from . import namespace as ns_
from .core_writer import write_defaults, write_fields, clean_fullname


def write_protocol_request(message, namespace, writer, use_logical_types):
    """
    Write concrete class for a protocol request
    :param avro.protocol.Message message: Message whose request to write
    :param TabbedWriter writer: writer to write to
    :param str namespace: protocol's namespace
    :return:
    """

    fullname = ns_.make_fullname(namespace, clean_fullname(message.name))
    namespace, type_name = ns_.split_fullname(fullname)

    writer.write('''\nclass {name}RequestClass(DictWrapper):'''.format(name=type_name))

    with writer.indent():
        writer.write("\n\n")

        writer.write('\nRECORD_SCHEMA = PROTOCOL_MESSAGES["%s"].request' % clean_fullname(message.name))

        writer.write('\n\n\ndef __init__(self, inner_dict=None):')
        with writer.indent():
            writer.write('\n')
            writer.write('super(RequestClasses.{name}RequestClass, self).__init__(inner_dict)'.format(name=fullname))

            writer.write('\nif inner_dict is None:')
            with writer.indent():
                write_defaults(message.request, writer, my_full_name=fullname + "Request", use_logical_types=use_logical_types)
        write_fields(message.request, writer, use_logical_types)



import six
import os

if six.PY3:
    from io import StringIO
else:
    from cStringIO import StringIO

from avro import protocol, schema

from . import namespace as ns_
from .tabbed_writer import TabbedWriter
from .core_writer import write_preamble, write_get_schema, start_namespace, write_reader_impl, clean_fullname
from .core_writer import write_schema_record, write_enum, write_read_file, generate_namespace_modules
from .protocol_writer import write_protocol_request


def generate_protocol(protocol_json, use_logical_types=False, custom_imports=None, avro_json_converter=None):
    """
    Generate content of the file which will contain concrete classes for RecordSchemas and requests contained
    in the avro protocol
    :param str protocol_json: JSON containing avro protocol
    :param bool use_logical_types: Use logical types extensions if true
    :param list[str] custom_imports: Add additional import modules
    :param str avro_json_converter: AvroJsonConverter type to use for default values
    :return:
    """

    if avro_json_converter is None:
        avro_json_converter = 'avrojson.AvroJsonConverter'

    if '(' not in avro_json_converter:
        avro_json_converter += '(use_logical_types=%s, schema_types=__SCHEMA_TYPES)' % use_logical_types

    custom_imports = custom_imports or []

    if not hasattr(protocol, 'parse'):
        # Older versions of avro used a capital P in Parse.
        proto = protocol.Parse(protocol_json)
    else:
        proto = protocol.parse(protocol_json)

    schemas = []
    messages = []
    schema_names = set()
    request_names = set()

    known_types = set()
    for schema_idx, record_schema in enumerate(proto.types):
        if isinstance(record_schema, (schema.RecordSchema, schema.EnumSchema)):
            schemas.append((schema_idx, record_schema))
            known_types.add(clean_fullname(record_schema.fullname))

    for message in (six.itervalues(proto.messages) if six.PY2 else proto.messages):
        messages.append((message, message.request, message.response if isinstance(message.response, (
            schema.EnumSchema, schema.RecordSchema)) and clean_fullname(message.response.fullname) not in known_types else None))
        if isinstance(message.response, (schema.EnumSchema, schema.RecordSchema)):
            known_types.add(clean_fullname(message.response.fullname))

    namespaces = {}
    for schema_idx, record_schema in schemas:
        ns, name = ns_.split_fullname(clean_fullname(record_schema.fullname))
        if ns not in namespaces:
            namespaces[ns] = {'requests': [], 'records': [], 'responses': []}
        namespaces[ns]['records'].append((schema_idx, record_schema))

    for message, request, response in messages:
        fullname = ns_.make_fullname(proto.namespace, clean_fullname(message.name))
        ns, name = ns_.split_fullname(fullname)
        if ns not in namespaces:
            namespaces[ns] = {'requests': [], 'records': [], 'responses': []}
        namespaces[ns]['requests'].append(message)
        if response:
            namespaces[ns]['responses'].append(message)

    main_out = StringIO()
    writer = TabbedWriter(main_out)

    write_preamble(writer, use_logical_types, custom_imports)
    write_protocol_preamble(writer, use_logical_types, custom_imports)
    write_get_schema(writer)
    write_populate_schemas(writer)

    writer.write('\n\n\nclass SchemaClasses(object):')
    with writer.indent():
        writer.write('\n\n')

        current_namespace = tuple()
        all_ns = sorted(namespaces.keys())

        for ns in all_ns:
            if not (namespaces[ns]['responses'] or namespaces[ns]['records']):
                continue

            namespace = ns.split('.')
            if namespace != current_namespace:
                start_namespace(current_namespace, namespace, writer)

            for idx, record in namespaces[ns]['records']:
                schema_names.add(clean_fullname(record.fullname))
                if isinstance(record, schema.RecordSchema):
                    write_schema_record(record, writer, use_logical_types)
                elif isinstance(record, schema.EnumSchema):
                    write_enum(record, writer)

            for message in namespaces[ns]['responses']:
                schema_names.add(clean_fullname(message.response.fullname))
                if isinstance(message.response, schema.RecordSchema):
                    write_schema_record(message.response, writer, use_logical_types)
                elif isinstance(message.response, schema.EnumSchema):
                    write_enum(message.response, writer)

        writer.write('\n\npass')

    writer.set_tab(0)
    writer.write('\n\n\nclass RequestClasses(object):')
    with writer.indent() as indent:
        writer.write('\n\n')

        current_namespace = tuple()
        all_ns = sorted(namespaces.keys())

        for ns in all_ns:
            if not (namespaces[ns]['requests'] or namespaces[ns]['responses']):
                continue

            namespace = ns.split('.')
            if namespace != current_namespace:
                start_namespace(current_namespace, namespace, writer)

            for message in namespaces[ns]['requests']:
                request_names.add(ns_.make_fullname(proto.namespace, clean_fullname(message.name)))
                write_protocol_request(message, proto.namespace, writer, use_logical_types)

        writer.write('\n\npass')

    writer.untab()
    writer.set_tab(0)
    writer.write('\n__SCHEMA_TYPES = {\n')
    writer.tab()

    all_ns = sorted(namespaces.keys())
    for ns in all_ns:
        for idx, record in (namespaces[ns]['records'] or []):
            writer.write("'%s': SchemaClasses.%sClass,\n" % (clean_fullname(record.fullname),
                                                             clean_fullname(record.fullname)))

        for message in (namespaces[ns]['responses'] or []):
            writer.write("'%s': SchemaClasses.%sClass,\n" % (clean_fullname(message.response.fullname),
                                                             clean_fullname(message.response.fullname)))

        for message in (namespaces[ns]['requests'] or []):
            name = ns_.make_fullname(proto.namespace, clean_fullname(message.name))
            writer.write("'%s': RequestClasses.%sRequestClass, \n" % (name, name))

    writer.untab()
    writer.write('\n}\n')

    writer.write('_json_converter = %s\n\n' % avro_json_converter)
    value = main_out.getvalue()
    main_out.close()
    return value, schema_names, request_names


def write_protocol_preamble(writer, use_logical_types, custom_imports):
    """
    Writes a preamble for avro protocol implementation.
    The preamble will contain a function which can load the protocol from the file
    and a global PROTOCOL variable which will contain parsed protocol
    :param writer:
    :param use_logical_types:
    :return:
    """
    write_read_file(writer)
    writer.write('\nfrom avro import protocol as avro_protocol')

    for i in (custom_imports or []):
        writer.write('import %s\n' % i)

    if use_logical_types:
        writer.write('\nfrom avrogen import logical')
    writer.write('\n\ndef __get_protocol(file_name):')
    with writer.indent():
        writer.write('\nproto = avro_protocol.Parse(__read_file(file_name)) if six.PY3 else avro_protocol.parse(__read_file(file_name))')
        writer.write('\nreturn proto')
    writer.write('\n\nPROTOCOL = __get_protocol(os.path.join(os.path.dirname(__file__), "protocol.avpr"))')


def write_populate_schemas(writer):
    """
    Write code which will look through the protocol and populate __SCHEMAS dict which will be used by get_type_schema()
    :param writer:
    :return:
    """
    writer.write('\nfor rec in PROTOCOL.types:')
    with writer.indent():
        writer.write('\n__SCHEMAS[rec.fullname] = rec')

    writer.write('\nfor resp in (six.itervalues(PROTOCOL.messages) if six.PY2 else PROTOCOL.messages):')
    with writer.indent():
        writer.write('\nif isinstance(resp.response, (avro_schema.RecordSchema, avro_schema.EnumSchema)):')
        with writer.indent():
            writer.write('\n__SCHEMAS[resp.response.fullname] = resp.response')

    writer.write('\nPROTOCOL_MESSAGES = {m.name.lstrip("."):m for m in (six.itervalues(PROTOCOL.messages) if six.PY2 else PROTOCOL.messages)}\n')


def write_protocol_files(protocol_json, output_folder, use_logical_types=False, custom_imports=None):
    """
    Generates concrete classes for RecordSchemas and requests and a SpecificReader for types and messages contained
    in the avro protocol.
    :param str protocol_json: JSON containing avro protocol
    :param str output_folder: Folder to write generated files to.
    :param list[str] custom_imports: Add additional import modules
    :return:
    """
    proto_py, record_names, request_names = generate_protocol(protocol_json, use_logical_types, custom_imports)
    names = sorted(list(record_names) + list(request_names))
    if not os.path.isdir(output_folder):
        os.mkdir(output_folder)

    with open(os.path.join(output_folder, "schema_classes.py"), "w+") as f:
        f.write(proto_py)

    with open(os.path.join(output_folder, "protocol.avpr"), "w+") as f:
        f.write(protocol_json)

    ns_dict = generate_namespace_modules(names, output_folder)

    with open(os.path.join(output_folder, "__init__.py"), "w+") as f:
        pass

    write_namespace_modules(ns_dict, request_names, output_folder)
    write_specific_reader(record_names, output_folder, use_logical_types)


def write_specific_reader(record_types, output_folder, use_logical_types):
    """
    Write specific reader implementation for a protocol
    :param list[avro.schema.RecordSchema] record_types:
    :param output_folder:
    :return:
    """
    with open(os.path.join(output_folder, "__init__.py"), "a+") as f:
        writer = TabbedWriter(f)
        writer.write('\n\nfrom .schema_classes import SchemaClasses, PROTOCOL as my_proto, get_schema_type')
        writer.write('\nfrom avro.io import DatumReader')

        write_reader_impl(record_types, writer, use_logical_types)


def write_namespace_modules(ns_dict, request_names, output_folder):
    """
    Writes content of the generated namespace modules. A python module will be created for each namespace
    and will import concrete schema classes from SchemaClasses
    :param ns_dict:
    :param request_names:
    :param output_folder:
    :return:
    """
    for ns in six.iterkeys(ns_dict):
        with open(os.path.join(output_folder, ns.replace('.', os.path.sep), "__init__.py"), "w+") as f:
            currency = '.'
            if ns != '':
                currency += '.' * len(ns.split('.'))
            f.write('from {currency}schema_classes import SchemaClasses\n'.format(currency=currency))
            f.write('from {currency}schema_classes import RequestClasses\n'.format(currency=currency))
            for name in ns_dict[ns]:
                if ns_.make_fullname(ns, name) in request_names:
                    f.write(
                        "{name}Request = RequestClasses.{ns}{name}RequestClass\n".format(name=name,
                                                                                         ns=ns if not ns else (
                                                                                         ns + ".")))
                else:
                    f.write("{name} = SchemaClasses.{ns}{name}Class\n".format(name=name,
                                                                              ns=ns if not ns else (ns + ".")))



from __future__ import division
from avro import schema, io
import functools
import abc
import six
import datetime
import decimal
import struct
import time
import pytz
import tzlocal

if six.PY3:
    long = int

EPOCH_DATE = datetime.date(1970, 1, 1)
SECONDS_IN_DAY = 24 * 60 * 60


class LogicalTypeProcessor(object, six.with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def validate(self, expected_schema, datum):
        pass

    @abc.abstractmethod
    def does_match(self, writers_schema, readers_schema):
        return False

    @abc.abstractmethod
    def convert_back(self, writers_schema, readers_schema, value):
        return value

    @abc.abstractmethod
    def convert(self, writers_schema, value):
        pass

    @abc.abstractmethod
    def can_convert(self, writers_schema):
        pass

    @abc.abstractmethod
    def typename(self):
        pass

    @abc.abstractmethod
    def initializer(self, value=None):
        pass


class DecimalLogicalTypeProcessor(LogicalTypeProcessor):
    def can_convert(self, writers_schema):
        return isinstance(writers_schema, schema.PrimitiveSchema) and writers_schema.type == 'string'

    def validate(self, expected_schema, datum):
        return isinstance(datum, (int, float, long, decimal.Decimal))

    def convert(self, writers_schema, value):
        if not isinstance(value, (int, float, long, decimal.Decimal)):
            raise Exception('Wrong type for decimal conversion')
        return str(value)

    def convert_back(self, writers_schema, readers_schema, value):
        return decimal.Decimal(value)

    def does_match(self, writers_schema, readers_schema):
        if isinstance(writers_schema, schema.PrimitiveSchema):
            if writers_schema.type == 'string':
                return True
        return False

    def typename(self):
        return 'decimal.Decimal'

    def initializer(self, value=None):
        return 'decimal.Decimal(%s)' % (0 if value is None else value)


class DateLogicalTypeProcessor(LogicalTypeProcessor):
    _matching_types = {'int', 'long', 'float', 'double'}

    def can_convert(self, writers_schema):
        return isinstance(writers_schema, schema.PrimitiveSchema) and writers_schema.type == 'int'

    def validate(self, expected_schema, datum):
        return isinstance(datum, datetime.date)

    def convert(self, writers_schema, value):
        if not isinstance(value, datetime.date):
            raise Exception("Wrong type for date conversion")
        return (value - EPOCH_DATE).total_seconds() // SECONDS_IN_DAY

    def convert_back(self, writers_schema, readers_schema, value):
        return EPOCH_DATE + datetime.timedelta(days=int(value))

    def does_match(self, writers_schema, readers_schema):
        if isinstance(writers_schema, schema.PrimitiveSchema):
            if writers_schema.type in DateLogicalTypeProcessor._matching_types:
                return True
        return False

    def typename(self):
        return 'datetime.date'

    def initializer(self, value=None):
        return ((
                    'logical.DateLogicalTypeProcessor().convert_back(None, None, %s)' % value) if value is not None
                else 'datetime.datetime.today().date()')


class TimeMicrosLogicalTypeProcessor(LogicalTypeProcessor):
    _matching_types = {'int', 'long', 'float', 'double'}

    def can_convert(self, writers_schema):
        return isinstance(writers_schema, schema.PrimitiveSchema) and writers_schema.type == 'long'

    def validate(self, expected_schema, datum):
        return isinstance(datum, datetime.time)

    def convert(self, writers_schema, value):
        if not isinstance(value, datetime.time):
            raise Exception('Wrong type for time conversion')
        return ((value.hour * 60 + value.minute) * 60 + value.second) * 1000000 + value.microsecond

    def convert_back(self, writers_schema, readers_schema, value):
        _, hours, minutes, seconds, microseconds = TimeMicrosLogicalTypeProcessor.extract_time_parts(value)
        return datetime.time(hours, minutes, seconds, microseconds)

    @staticmethod
    def extract_time_parts(value):
        value = long(value)
        days = value // 86400000000
        hours = value % 86400000000
        minutes = hours % 3600000000
        hours = hours // 3600000000
        seconds = minutes % 60000000
        minutes = minutes // 60000000
        microseconds = seconds % 1000000
        seconds = seconds // 1000000
        return days, hours, minutes, seconds, microseconds

    def does_match(self, writers_schema, readers_schema):
        if isinstance(writers_schema, schema.PrimitiveSchema):
            if writers_schema.type in TimeMicrosLogicalTypeProcessor._matching_types:
                return True
        return False

    def typename(self):
        return 'datetime.time'

    def initializer(self, value=None):
        return ((
                    'logical.TimeMicrosLogicalTypeProcessor().convert_back(None, None, %s)' % value) if value is not None
                else 'datetime.datetime.today().time()')


class TimeMillisLogicalTypeProcessor(TimeMicrosLogicalTypeProcessor):
    def can_convert(self, writers_schema):
        return isinstance(writers_schema, schema.PrimitiveSchema) and writers_schema.type == 'int'

    def convert(self, writers_schema, value):
        if not isinstance(value, datetime.time):
            raise Exception('Wrong type for time conversion')
        return int(super(TimeMillisLogicalTypeProcessor, self).convert(writers_schema, value) // 1000)

    def convert_back(self, writers_schema, readers_schema, value):
        return super(TimeMillisLogicalTypeProcessor, self).convert_back(writers_schema, readers_schema, value * 1000)

    def initializer(self, value=None):
        return ((
                    'logical.TimeMillisLogicalTypeProcessor().convert_back(None, None, %s)' % value) if value is not None
                else 'datetime.datetime.today().time()')


class TimestampMicrosLogicalTypeProcessor(LogicalTypeProcessor):
    _matching_types = {'int', 'long', 'float', 'double'}

    def can_convert(self, writers_schema):
        return isinstance(writers_schema, schema.PrimitiveSchema) and writers_schema.type == 'long'

    def validate(self, expected_schema, datum):
        return isinstance(datum, datetime.datetime)

    def convert(self, writers_schema, value):
        # below doesn't work for time zones which are ahead of UTC,
        # so will need to switch to datetime arithmetic
        #EPOCH_TT = time.mktime((1970, 1, 1, 0, 0, 0, 0, 0, 0))

        EPOCH_TT = -datetime.datetime(1970, 1, 1).astimezone(tzlocal.get_localzone()).utcoffset().total_seconds()

        if not isinstance(value, datetime.datetime):
            if isinstance(value, datetime.date):
                value = tzlocal.get_localzone().localize(
                    datetime.datetime(value.year, value.month, value.day, 0, 0, 0, 0))

        if value.tzinfo is None:
            value = tzlocal.get_localzone().localize(value)
        value = (time.mktime(value.utctimetuple()) - EPOCH_TT) + value.microsecond / 1000000.0
        return long(value * 1000000)

    def convert_back(self, writers_schema, readers_schema, value):
        value = long(value) / 1000000.0
        utc = datetime.datetime.utcfromtimestamp(value).replace(tzinfo=pytz.UTC)
        return utc.astimezone(tzlocal.get_localzone()).replace(tzinfo=None)

    def does_match(self, writers_schema, readers_schema):
        if isinstance(writers_schema, schema.PrimitiveSchema):
            if writers_schema.type in TimestampMicrosLogicalTypeProcessor._matching_types:
                return True
        return False

    def typename(self):
        return 'datetime.datetime'

    def initializer(self, value=None):
        return ((
                    'logical.TimestampMicrosLogicalTypeProcessor().convert_back(None, None, %s)' % value) if value is not None
                else 'datetime.datetime.now()')


class TimestampMillisLogicalTypeProcessor(TimestampMicrosLogicalTypeProcessor):
    def convert(self, writers_schema, value):
        return super(TimestampMillisLogicalTypeProcessor, self).convert(writers_schema, value) // 1000

    def convert_back(self, writers_schema, readers_schema, value):
        return super(TimestampMillisLogicalTypeProcessor, self).convert_back(writers_schema, readers_schema,
                                                                             value * 1000)

    def initializer(self, value=None):
        return ((
                    'logical.TimestampMillisLogicalTypeProcessor().convert_back(None, None, %s)' % value) if value is not None
                else 'datetime.datetime.now()')


DEFAULT_LOGICAL_TYPES = {
    'decimal': DecimalLogicalTypeProcessor(),
    'date': DateLogicalTypeProcessor(),
    'time-millis': TimeMillisLogicalTypeProcessor(),
    'time-micros': TimeMicrosLogicalTypeProcessor(),
    'timestamp-millis': TimestampMillisLogicalTypeProcessor(),
    'timestamp-micros': TimestampMicrosLogicalTypeProcessor(),
}


class LogicalDatumReader(io.DatumReader):
    def __init__(self, writers_schema=None, readers_schema=None, logical_types=DEFAULT_LOGICAL_TYPES):
        """
        Initializes DatumReader with logical type support

        :param schema.Schema writers_schema: Writer's schema
        :param schema.Schema readers_schema: Optional reader's schema
        :param dict[str, LogicalTypeProcessor] logical_types: Optional logical types dict
        """
        super(LogicalDatumReader, self).__init__(writers_schema=writers_schema, readers_schema=readers_schema)
        self.logical_types = logical_types or {}

    def read_data(self, writers_schema, readers_schema, decoder):
        """
        Reads data from the decoded stream and applies logical type conversion

        :param schema.Schema writers_schema:
        :param schema.Schema readers_schema:
        :param io.BinaryDecoder decoder:
        :return:
        """
        result = super(LogicalDatumReader, self).read_data(writers_schema, readers_schema, decoder)
        logical_type = readers_schema.props.get('logicalType')
        if logical_type:
            logical_type_handler = self.logical_types.get(logical_type)
            if logical_type_handler and logical_type_handler.does_match(writers_schema, readers_schema):
                result = logical_type_handler.convert_back(writers_schema, readers_schema, result)
        return result


class LogicalDatumWriter(io.DatumWriter):
    """
       Initializes DatumWriter with logical type support

       :param schema.Schema writers_schema: Writer's schema
       :param dict[str, LogicalTypeProcessor] logical_types: Optional logical types dict
       """

    def __init__(self, writers_schema=None, logical_types=DEFAULT_LOGICAL_TYPES):
        super(LogicalDatumWriter, self).__init__(writers_schema=writers_schema)
        self.logical_types = logical_types

    def write_data(self, writers_schema, datum, encoder):
        logical_type = writers_schema.props.get('logicalType')
        if logical_type:
            logical_type_handler = self.logical_types.get(logical_type)
            if logical_type_handler and logical_type_handler.can_convert(writers_schema):
                return super(LogicalDatumWriter, self).write_data(writers_schema,
                                                                  logical_type_handler.convert(writers_schema, datum),
                                                                  encoder)
        return super(LogicalDatumWriter, self).write_data(writers_schema, datum, encoder)

    def __validate(self, writers_schema, datum):
        logical_type = writers_schema.props.get('logicalType')
        if logical_type:
            lt = self.logical_types.get(logical_type)
            if lt:
                if lt.can_convert(writers_schema):
                    if lt.validate(writers_schema, datum):
                        return True
                    return False

        schema_type = writers_schema.type
        if schema_type == 'array':
            return (isinstance(datum, list) and
                    False not in [self.__validate(writers_schema.items, d) for d in datum])
        elif schema_type == 'map':
            return (isinstance(datum, dict) and
                    False not in [isinstance(k, basestring) for k in datum.keys()] and
                    False not in
                    [self.__validate(writers_schema.values, v) for v in datum.values()])
        elif schema_type in ['union', 'error_union']:
            return True in [self.__validate(s, datum) for s in writers_schema.schemas]
        elif schema_type in ['record', 'error', 'request']:
            return (isinstance(datum, dict) and
                    False not in
                    [self.__validate(f.type, datum.get(f.name)) for f in writers_schema.fields])

        return io.validate(writers_schema, datum)

    def write(self, datum, encoder):
        # validate datum
        if not self.__validate(self.writers_schema, datum):
            raise io.AvroTypeException(self.writers_schema, datum)

        self.write_data(self.writers_schema, datum, encoder)


def patch_logical_types():
    io.DatumWriter = LogicalDatumWriter
    io.DatumReader = LogicalDatumReader




import json
import os
import six
from avro import schema

from io import StringIO


from avro.schema import SchemaFromJSONData as make_avsc_object

from .core_writer import generate_namespace_modules, clean_fullname
from .tabbed_writer import TabbedWriter
from .core_writer import write_preamble, start_namespace, write_schema_record, write_enum, write_read_file
from .core_writer import write_get_schema, write_reader_impl
import logging

logger = logging.getLogger('avrogen.schema')
logger.setLevel(logging.INFO)


def generate_schema(schema_json, use_logical_types=False, custom_imports=None, avro_json_converter=None):
    """
    Generate file containing concrete classes for RecordSchemas in given avro schema json
    :param str schema_json: JSON representing avro schema
    :param list[str] custom_imports: Add additional import modules
    :param str avro_json_converter: AvroJsonConverter type to use for default values
    :return Dict[str, str]:
    """

    if avro_json_converter is None:
        avro_json_converter = 'avrojson.AvroJsonConverter'

    if '(' not in avro_json_converter:
        avro_json_converter += f'(use_logical_types={use_logical_types}, schema_types=__SCHEMA_TYPES)'

    custom_imports = custom_imports or []
    names = schema.Names()
    make_avsc_object(json.loads(schema_json), names)

    names = [k for k in six.iteritems(names.names) if isinstance(k[1], (schema.RecordSchema, schema.EnumSchema))]
    names = sorted(names, key=lambda x: x[0])

    main_out = StringIO()
    writer = TabbedWriter(main_out)

    write_preamble(writer, use_logical_types, custom_imports)
    write_schema_preamble(writer)
    write_get_schema(writer)
    write_populate_schemas(writer)

    current_namespace = tuple()

    for name, field_schema in names:  # type: str, schema.Schema
        name = clean_fullname(name)
        namespace = tuple(name.split('.')[:-1])
        if namespace != current_namespace:
            current_namespace = namespace
        if isinstance(field_schema, schema.RecordSchema):
            logger.debug(f'Writing schema: {clean_fullname(field_schema.fullname)}')
            write_schema_record(field_schema, writer, use_logical_types)
        elif isinstance(field_schema, schema.EnumSchema):
            logger.debug(f'Writing enum: {field_schema.fullname}', field_schema.fullname)
            write_enum(field_schema, writer)
    writer.set_tab(0)
    writer.write('\n__SCHEMA_TYPES = {')
    writer.tab()

    # Lookup table for fullname.
    for name, field_schema in names:
        n = clean_fullname(field_schema.name)
        full = field_schema.fullname
        writer.write(f"\n'{full}': {n}Class,")

    # Lookup table for names without namespace.
    for name, field_schema in names:
        n = clean_fullname(field_schema.name)
        writer.write(f"\n'{n}': {n}Class,")

    writer.untab()
    writer.write('\n}\n\n')

    writer.write(f'_json_converter = {avro_json_converter}\n\n')

    value = main_out.getvalue()
    main_out.close()
    return value, [clean_fullname(name[0]) for name in names]


def write_schema_preamble(writer):
    """
    Writes a schema-specific preamble: __get_names_and_schema() which is used by concrete classes to resolve
    their own RecordSchema
    :param writer:
    :return:
    """
    write_read_file(writer)
    writer.write('\n\ndef __get_names_and_schema(json_str):')
    with writer.indent():
        writer.write('\nnames = avro_schema.Names()')
        writer.write('\nschema = make_avsc_object(json.loads(json_str), names)')
        writer.write('\nreturn names, schema')
    writer.write('\n\n\nSCHEMA_JSON_STR = __read_file(os.path.join(os.path.dirname(__file__), "schema.avsc"))')
    writer.write('\n\n\n__NAMES, SCHEMA = __get_names_and_schema(SCHEMA_JSON_STR)')


def write_populate_schemas(writer):
    """
    Writes out a __SCHEMAS dict which contains all RecordSchemas by their full name. Used by get_schema_type
    :param writer:
    :return:
    """
    writer.write('\n__SCHEMAS = dict((n.fullname.lstrip("."), n) for n in six.itervalues(__NAMES.names))\n')


def write_namespace_modules(ns_dict, output_folder):
    """
    Writes content of the generated namespace modules. A python module will be created for each namespace
    and will import concrete schema classes from SchemaClasses
    :param ns_dict:
    :param output_folder:
    :return:
    """
    for ns in six.iterkeys(ns_dict):
        with open(os.path.join(output_folder, ns.replace('.', os.path.sep), "__init__.py"), "w+") as f:
            currency = '.'
            if ns != '':
                currency += '.' * len(ns.split('.'))
            for name in ns_dict[ns]:
                f.write(f'from {currency}schema_classes import {name}Class\n')

            f.write('\n\n')

            for name in ns_dict[ns]:
                f.write(f"{name} = {name}Class\n")


def write_specific_reader(record_types, output_folder, use_logical_types):
    """
    Writes specific reader for a avro schema into generated root module
    :param record_types:
    :param output_folder:
    :return:
    """
    with open(os.path.join(output_folder, "__init__.py"), "a+") as f:
        writer = TabbedWriter(f)
        writer.write('from typing import cast')
        writer.write('\nfrom avrogen.dict_wrapper import DictWrapper')
        writer.write('\nfrom .schema_classes import SCHEMA as get_schema_type')
        writer.write('\nfrom .schema_classes import _json_converter as json_converter')
        for t in record_types:
            writer.write(f'\nfrom .schema_classes import {t.split(".")[-1]}Class')
        writer.write('\nfrom avro.io import DatumReader')
        if use_logical_types:
            writer.write('\nfrom avrogen import logical')

        write_reader_impl(record_types, writer, use_logical_types)


def write_schema_files(schema_json, output_folder, use_logical_types=False, custom_imports=None):
    """
    Generates concrete classes, namespace modules, and a SpecificRecordReader for a given avro schema
    :param str schema_json: JSON containing avro schema
    :param str output_folder: Folder in which to create generated files
    :param list[str] custom_imports: Add additional import modules
    :return:
    """
    schema_py, names = generate_schema(schema_json, use_logical_types, custom_imports)
    names = sorted(names)

    if not os.path.isdir(output_folder):
        os.mkdir(output_folder)

    with open(os.path.join(output_folder, "schema_classes.py"), "w+") as f:
        f.write(schema_py)

    with open(os.path.join(output_folder, "schema.avsc"), "w+") as f:
        f.write(schema_json)

    ns_dict = generate_namespace_modules(names, output_folder)

    with open(os.path.join(output_folder, "__init__.py"), "w+") as f:
        pass  # make sure we create this file from scratch

    write_namespace_modules(ns_dict, output_folder)
    write_specific_reader(names, output_folder, use_logical_types)



"""Usage:
  python -m avrogen /path/to/protocol.avdl [-o /path/to/output]
"""

from .protocol import write_protocol_files
from os.path import join
from sys import argv

def main():

	if "-o" in argv:
		avdl, _, output = argv[1:]
	else:
		avdl, output = argv[1], './'
	write_protocol_files(open(avdl, 'r').read(), output)

if __name__ == '__main__':
	try:
		exit(main())
	except Exception:
		print(__doc__)
		raise


import os

from avro import schema
from . import namespace as ns_
from . import logical
import six
import keyword

if six.PY3:
    long = int


PRIMITIVE_TYPES = {
    'null',
    'boolean',
    'int',
    'long',
    'float',
    'double',
    'bytes',
    'string'
}
__PRIMITIVE_TYPE_MAPPING = {
    'null': None,
    'boolean': bool,
    'int': int,
    'long': long,
    'float': float,
    'double': float,
    'bytes': bytes,
    'string': str,
}

def clean_fullname(fullname):
    if six.PY3:
        return fullname.lstrip('.')
    return fullname


def convert_default(full_name, idx, do_json=True):
    if do_json:
        return (f'_json_converter.from_json_object({full_name}Class.RECORD_SCHEMA.fields[{idx}].default,'
               + f' writers_schema={full_name}Class.RECORD_SCHEMA.fields[{idx}].type)')
    else:
        return f'self.RECORD_SCHEMA.field_map["{idx}"].default'


def get_default(field, use_logical_types, my_full_name=None, f_name=None):
    default_written = False
    f_name = f_name if f_name is not None else field.name
    if keyword.iskeyword(field.name):
        f_name =  field.name + get_field_type_name(field.type, use_logical_types)

    default_type, nullable = find_type_of_default(field.type)
    if field.has_default:
        if use_logical_types and default_type.props.get('logicalType') \
                and default_type.props.get('logicalType') in logical.DEFAULT_LOGICAL_TYPES:
            lt = logical.DEFAULT_LOGICAL_TYPES[default_type.props.get('logicalType')]
            v = lt.initializer(convert_default(my_full_name,
                                                idx=i,
                                                do_json=isinstance(default_type,
                                                schema.RecordSchema)))
            return v
        elif isinstance(default_type, schema.RecordSchema):
            d = convert_default(idx=i, full_name=my_full_name, do_json=True)
            return f'{field.name}Class({d})'
        elif isinstance(default_type, (schema.PrimitiveSchema, schema.EnumSchema, schema.FixedSchema)):
            d = convert_default(full_name=my_full_name, idx=f_name, do_json=False)
            return d

    if not default_written:
        default_written = True
        if nullable:
            return 'None'
        elif use_logical_types and default_type.props.get('logicalType') \
                and default_type.props.get('logicalType') in logical.DEFAULT_LOGICAL_TYPES:
            lt = logical.DEFAULT_LOGICAL_TYPES[default_type.props.get('logicalType')]
            return f'{default}'.format(default=lt.initializer())
        elif isinstance(default_type, schema.PrimitiveSchema) and not default_type.props.get('logicalType'):
            d = get_primitive_field_initializer(default_type)
            return d
        elif isinstance(default_type, schema.EnumSchema):
            f = clean_fullname(default_type.name)
            s = default_type.symbols[0]
            return f'{f}Class.{s}'
        elif isinstance(default_type, schema.MapSchema):
            return 'dict()'
        elif isinstance(default_type, schema.ArraySchema):
            return 'list()'
        elif isinstance(default_type, schema.FixedSchema):
            return 'bytes()'
        elif isinstance(default_type, schema.RecordSchema):
            f = clean_fullname(default_type.name)
            return f'{f}Class.construct_with_defaults()'
    raise AttributeError('cannot get default for field')

def write_defaults(record, writer, my_full_name=None, use_logical_types=False):
    """
    Write concrete record class's constructor part which initializes fields with default values
    :param schema.RecordSchema record: Avro RecordSchema whose class we are generating
    :param TabbedWriter writer: Writer to write to
    :param str my_full_name: Full name of the RecordSchema we are writing. Should only be provided for protocol requests.
    :return:
    """
    i = 0
    my_full_name = my_full_name or clean_fullname(record.fullname)

    something_written = False
    for field in record.fields:
        f_name = field.name
        if keyword.iskeyword(field.name):
            f_name =  field.name + get_field_type_name(field.type, use_logical_types)
        default = get_default(field, use_logical_types, my_full_name=my_full_name, f_name=f_name)
        writer.write(f'\nself.{f_name} = {default}')
        something_written = True
        i += 1
    if not something_written:
        writer.write('\npass')


def write_fields(record, writer, use_logical_types):
    """
    Write field definitions for a given RecordSchema
    :param schema.RecordSchema record: Avro RecordSchema we are generating
    :param TabbedWriter writer: Writer to write to
    :return:
    """
    writer.write('\n\n')
    for field in record.fields:  # type: schema.Field
        write_field(field, writer, use_logical_types)

def get_field_name(field, use_logical_types):
    name = field.name
    if keyword.iskeyword(field.name):
        name =  field.name + get_field_type_name(field.type, use_logical_types)
    return name

def write_field(field, writer, use_logical_types):
    """
    Write a single field definition
    :param field:
    :param writer:
    :return:
    """
    name = get_field_name(field, use_logical_types)
    doc = field.doc
    get_docstring = f'"""Getter: {doc}"""' if doc else "# No docs available."
    set_docstring = f'"""Setter: {doc}"""' if doc else "# No docs available."
    writer.write('''
@property
def {name}(self) -> {ret_type_name}:
    {get_docstring}
    return self._inner_dict.get('{raw_name}')  # type: ignore


@{name}.setter
def {name}(self, value: {ret_type_name}) -> None:
    {set_docstring}
    self._inner_dict['{raw_name}'] = value

'''.format(name=name, get_docstring=get_docstring, set_docstring=set_docstring, raw_name=field.name, ret_type_name=get_field_type_name(field.type, use_logical_types)))


def get_primitive_field_initializer(field_schema):
    """
    Gets a python code string which represents a type initializer for a primitive field.
    Used for required fields where no default is provided. Output will look like "int()" or similar
    :param schema.PrimitiveSchema field_schema:
    :return:
    """

    if field_schema.type == 'null':
        return 'None'
    return get_field_type_name(field_schema, False) + "()"


def get_field_type_name(field_schema, use_logical_types):
    """
    Gets a python type-hint for a given schema
    :param schema.Schema field_schema:
    :return: String containing python type hint
    """
    if use_logical_types and field_schema.props.get('logicalType'):
        from avrogen.logical import DEFAULT_LOGICAL_TYPES
        lt = DEFAULT_LOGICAL_TYPES.get(field_schema.props.get('logicalType'))
        if lt:
            return lt.typename()

    if isinstance(field_schema, schema.PrimitiveSchema):
        if field_schema.fullname == 'null':
            return 'None'
        return __PRIMITIVE_TYPE_MAPPING[field_schema.fullname].__name__
    elif isinstance(field_schema, schema.FixedSchema):
        return 'bytes'
    elif isinstance(field_schema, schema.EnumSchema):
        # For enums, we have their "class" types, but they're actually
        # represented as strings. This is a decent hack to work around
        # the issue.
        return f'Union[str, "{field_schema.name}Class"]'
    elif isinstance(field_schema, schema.NamedSchema):
        return f'"{field_schema.name}Class"'
    elif isinstance(field_schema, schema.ArraySchema):
        return 'List[' + get_field_type_name(field_schema.items, use_logical_types) + ']'
    elif isinstance(field_schema, schema.MapSchema):
        return 'Dict[str, ' + get_field_type_name(field_schema.values, use_logical_types) + ']'
    elif isinstance(field_schema, schema.UnionSchema):
        type_names = [get_field_type_name(x, use_logical_types) for x in field_schema.schemas if
                      get_field_type_name(x, use_logical_types)]
        if len(type_names) > 1:
            return 'Union[' + ', '.join(type_names) + ']'
        elif len(type_names) == 1:
            return type_names[0]
        return ''


def find_type_of_default(field_type):
    """
    Returns full name of an avro type of the field's default value
    :param schema.Schema field_type:
    :return:
    """

    if isinstance(field_type, schema.UnionSchema):
        # For union types, the default is always the first item.
        field, nullable = find_type_of_default(field_type.schemas[0])
        return field, nullable
        # non_null_types = [s for s in field_type.schemas if s.type != 'null']
        # if non_null_types:
        #     type_, nullable = find_type_of_default(non_null_types[0])
        #     nullable = nullable or any(
        #         f for f in field_type.schemas if isinstance(f, schema.PrimitiveSchema) and f.fullname == 'null')
        # else:
        #     type_, nullable = field_type.schemas[0], True
        # return type_, nullable
    elif isinstance(field_type, schema.PrimitiveSchema):
        return field_type, field_type.fullname == 'null'
    else:
        return field_type, False


def start_namespace(current, target, writer):
    """
    Writes a new class corresponding to the target namespace to the schema file and
     closes the prior namespace
    :param tuple[str] current: Current namespace
    :param tuple[str] target: Target namespace we need to generate classes for
    :param TabbedWriter writer:
    :return:
    """

    i = 0
    while i < min(len(current), len(target)) and current[i] == target[i]:
        i += 1

    writer.write('\n\n')
    writer.set_tab(0)
    writer.write('\n')
    for component in target[i:]:
        writer.write('class {name}(object):'.format(name=component))
        writer.tab()
        writer.write('\n')


def write_preamble(writer, use_logical_types, custom_imports):
    """
    Writes a preamble of the file containing schema classes
    :param  writer:
    :return:
    """
    writer.write('import json\n')
    writer.write('import os.path\n')
    writer.write('import decimal\n')
    writer.write('import datetime\n')
    writer.write('import six\n')

    for cs in (custom_imports or []):
        writer.write(f'import {cs}\n')
    writer.write('from avrogen.dict_wrapper import DictWrapper\n')
    writer.write('from avrogen import avrojson\n')
    if use_logical_types:
        writer.write('from avrogen import logical\n')
    writer.write('from avro.schema import RecordSchema, SchemaFromJSONData as make_avsc_object\n')
    writer.write('from avro import schema as avro_schema\n')
    writer.write('from typing import List, Dict, Union, Optional, overload\n')
    writer.write('\n')


def write_read_file(writer):
    """
    Write a function which reads our schema or protocol
    :param writer:
    :return:
    """
    writer.write('\ndef __read_file(file_name):')
    with writer.indent():
        writer.write('\nwith open(file_name, "r") as f:')
        with writer.indent():
            writer.write('\nreturn f.read()\n')


def write_get_schema(writer):
    """
    Write get_schema_type which is used by concrete classes to resolve their own RecordSchemas
    :param writer:
    :return:
    """
    writer.write('\n__SCHEMAS: Dict[str, RecordSchema] = {}\n\n\n')
    writer.write('def get_schema_type(fullname):')
    with writer.indent():
        writer.write('\nreturn __SCHEMAS.get(fullname)\n\n')


def write_reader_impl(record_types, writer, use_logical_types):
    """
    Write specific reader implementation
    :param list[schema.RecordSchema] record_types:
    :param writer:
    :return:
    """
    writer.write('\n\n\nclass SpecificDatumReader(%s):' % (
        'DatumReader' if not use_logical_types else 'logical.LogicalDatumReader'))
    with writer.indent():
        writer.write('\nSCHEMA_TYPES = {')
        with writer.indent():
            for t in record_types:
                t_class = t.split('.')[-1]
                writer.write('\n"{t_class}": {t_class}Class,'.format(t_class=t_class))
                writer.write('\n".{t_class}": {t_class}Class,'.format(t_class=t_class))
                writer.write('\n"{f_class}": {t_class}Class,'.format(t_class=t_class, f_class=t))

        writer.write('\n}')
        writer.write('\n\n\ndef __init__(self, readers_schema=None, **kwargs):')
        with writer.indent():
            writer.write('\nwriters_schema = kwargs.pop("writers_schema", readers_schema)')
            writer.write('\nwriters_schema = kwargs.pop("writer_schema", writers_schema)')
            writer.write('\nsuper(SpecificDatumReader, self).__init__(writers_schema, readers_schema, **kwargs)')

        writer.write('\n\n\ndef read_record(self, writers_schema, readers_schema, decoder):')
        with writer.indent():
            writer.write(
                '\nresult = super(SpecificDatumReader, self).read_record(writers_schema, readers_schema, decoder)')
            writer.write('\n\nif readers_schema.fullname in SpecificDatumReader.SCHEMA_TYPES:')
            with writer.indent():
                writer.write('\ntp = SpecificDatumReader.SCHEMA_TYPES[readers_schema.fullname]')
                writer.write('\nif issubclass(tp, DictWrapper):')
                writer.write('\n    result = tp.construct(result)')
                writer.write('\nelse:')
                writer.write('\n    # tp is an enum')
                writer.write('\n    result = tp(result)  # type: ignore')
            writer.write('\n\nreturn result')


def generate_namespace_modules(names, output_folder):
    """
    Generate python modules corresponding to schema/protocol namespaces.

    :param names:
    :param output_folder:
    :return: Dictinoary of (namespace, list(name))
    :rtype: dict[str, list[str]]
    """
    ns_dict = {}
    for name in names:
        name_parts = name.split('.')
        full_path = output_folder
        for part in name_parts[:-1]:
            full_path = os.path.join(full_path, part)
            if not os.path.isdir(full_path):
                os.mkdir(full_path)
                # make sure __init__.py is created for every namespace level
                with open(os.path.join(full_path, "__init__.py"), "w+"): pass

        ns = '.'.join(name_parts[:-1])
        if not ns in ns_dict:
            ns_dict[ns] = []
        ns_dict[ns].append(name_parts[-1])
    return ns_dict


def write_schema_record(record, writer, use_logical_types):
    """
    Writes class representing Avro record schema
    :param avro.schema.RecordSchema record:
    :param TabbedWriter writer:
    :return:
    """

    _, type_name = ns_.split_fullname(record.fullname)
    writer.write('''\nclass {name}Class(DictWrapper):'''.format(name=type_name))

    with writer.indent():
        writer.write('\n')
        if record.doc:
            writer.write(f'"""{record.doc}"""')
        else:
            writer.write('# No docs available.')
        writer.write('\n\nRECORD_SCHEMA = get_schema_type("%s")' % (record.namespace + '.' + record.name))

        write_record_init(record, writer, use_logical_types)

        write_serialization_stubs(record, writer, use_logical_types)

        write_fields(record, writer, use_logical_types)


def write_record_init(record, writer, use_logical_types):
    writer.write('\ndef __init__(self,')
    with writer.indent():
        delayed_lines = []
        for field in record.fields:  # type: schema.Field
            name = get_field_name(field, use_logical_types)
            ret_type_name = get_field_type_name(field.type, use_logical_types)
            default_type, nullable = find_type_of_default(field.type)

            if nullable:
                delayed_lines.append(f'\n{name}: {ret_type_name}=None,')
            else:
                writer.write(f'\n{name}: {ret_type_name},')
            # default = get_default(field, use_logical_types)
            # writer.write(f'\n{name}: {ret_type_name} = {default},')
        for line in delayed_lines:
            writer.write(line)
    writer.write('\n):')
    with writer.indent():
        writer.write('\n')
        writer.write('super().__init__()')
        writer.write('\n')

        for field in record.fields:  # type: schema.Field
            name = get_field_name(field, use_logical_types)
            writer.write(f'\nself.{name} = {name}')

    writer.write('\n\n@classmethod')
    writer.write(f'\ndef construct_with_defaults(cls) -> "{record.name}Class":')
    with writer.indent():
        writer.write('\nself = cls.construct({})')
        writer.write('\nself._restore_defaults()')
        writer.write('\n')
        writer.write('\nreturn self')

    writer.write('\n')
    writer.write(f'\ndef _restore_defaults(self) -> None:')
    with writer.indent():
        write_defaults(record, writer, use_logical_types=use_logical_types)


def write_serialization_stubs(record, writer, use_logical_types):
    # writer.write(f'\n\n@classmethod')
    # writer.write(f'\ndef from_obj(cls, obj: dict, tuples=False) -> "{record.name}Class":')
    # with writer.indent():
    #     writer.write('\n...')

    # writer.write(f'\n\ndef to_obj(self, tuples=False) -> dict:')
    # with writer.indent():
    #     writer.write('\n...')
    pass


def write_enum(enum, writer):
    """
    Write class representing Avro enum schema
    :param schema.EnumSchema enum:
    :param TabbedWriter writer:
    :return:
    """
    _, type_name = ns_.split_fullname(enum.fullname)
    writer.write('''\nclass {name}Class(object):'''.format(name=type_name))

    with writer.indent():
        writer.write('\n')
        if enum.doc:
            writer.write(f'"""{enum.doc}"""')
        else:
            writer.write('# No docs available.')

        writer.write('\n\n')
        for field in enum.symbols:
            writer.write('{name} = "{name}"\n'.format(name=field))
        writer.write('\n')



import collections
import six

from . import logical
from .dict_wrapper import DictWrapper
from avro import schema
from avro import io

if six.PY3:
    io_validate = io.Validate
else:
    io_validate = io.validate

_PRIMITIVE_TYPES = set(schema.PRIMITIVE_TYPES)


class AvroJsonConverter(object):
    def __init__(self, use_logical_types=False, logical_types=logical.DEFAULT_LOGICAL_TYPES, schema_types=None):
        self.use_logical_types = use_logical_types
        self.logical_types = logical_types or {}
        self.schema_types = schema_types or {}
        self.fastavro = False
        
        # Register self with all the schema objects.
        for klass in self.schema_types.values():
            klass._json_converter = self
    
    def with_tuple_union(self, enable=True) -> 'AvroJsonConverter':
        ret = AvroJsonConverter(self.use_logical_types, self.logical_types, self.schema_types)
        ret.fastavro = enable
        return ret

    def validate(self, expected_schema, datum, skip_logical_types=False):
        if self.use_logical_types and expected_schema.props.get('logicalType') and not skip_logical_types \
                and expected_schema.props.get('logicalType') in self.logical_types:
            return self.logical_types[expected_schema.props.get('logicalType')].can_convert(expected_schema) \
                   and self.logical_types[expected_schema.props.get('logicalType')].validate(expected_schema, datum)
        schema_type = expected_schema.type
        if schema_type == 'array':
            return (isinstance(datum, list) and
                    False not in [self.validate(expected_schema.items, d, skip_logical_types) for d in datum])
        elif schema_type == 'map':
            return (isinstance(datum, dict) and
                    False not in [isinstance(k, six.string_types) for k in datum.keys()] and
                    False not in
                    [self.validate(expected_schema.values, v, skip_logical_types) for v in datum.values()])
        elif schema_type in ['union', 'error_union']:
            if isinstance(datum, DictWrapper):
                # Match the type based on the declared schema.
                data_schema = self._get_record_schema_if_available(datum)
                for i, candidate_schema in enumerate(expected_schema.schemas):
                    if candidate_schema.namespace == data_schema.namespace \
                        and candidate_schema.name == data_schema.name:
                        return self.validate(candidate_schema, datum)

            # If the union type is using a "name" to distinguish the type, we
            # must handle this specially during validation.
            value_type = None
            if not self.fastavro and isinstance(datum, dict):
                if len(datum) == 1:
                    items = list(six.iteritems(datum))
                    if not items:
                        return None
                    value_type = items[0][0]
                    value = items[0][1]
            elif self.fastavro and (isinstance(datum, list) or isinstance(datum, tuple)):
                if len(datum) == 2:
                    value_type = datum[0]
                    value = datum[1]
            if value_type is not None:
                for s in expected_schema.schemas:
                    name = self._fullname(s)
                    if name == value_type:
                        if self.validate(s, value, skip_logical_types):
                            return True
                        # If the specialized validation fails, we still attempt normal validation.

            return True in [self.validate(s, datum, skip_logical_types) for s in expected_schema.schemas]
        elif schema_type in ['record', 'error', 'request']:
            return ((isinstance(datum, dict) or isinstance(datum, DictWrapper)) and
                    False not in
                    [self.validate(f.type, datum.get(f.name), skip_logical_types) for f in expected_schema.fields])

        return io_validate(expected_schema, datum)

    def from_json_object(self, json_obj, writers_schema=None, readers_schema=None):
        if readers_schema is None:
            readers_schema = writers_schema
        if writers_schema is None:
            writers_schema = readers_schema

        if writers_schema is None:
            raise Exception('At least one schema must be specified')

        if not io.DatumReader.match_schemas(writers_schema, readers_schema):
            raise io.SchemaResolutionException('Could not match schemas', writers_schema, readers_schema)

        return self._generic_from_json(json_obj, writers_schema, readers_schema)

    def to_json_object(self, data_obj, writers_schema=None):
        if writers_schema is None:
            writers_schema = self._get_record_schema_if_available(data_obj)

        if writers_schema is None:
            raise Exception("Could not determine writer's schema from the object type and schema was not passed")
        assert isinstance(writers_schema, schema.Schema)

        if not self.validate(writers_schema, data_obj):
            raise io.AvroTypeException(writers_schema, data_obj)

        return self._generic_to_json(data_obj, writers_schema)

    def _fullname(self, schema_):
        if isinstance(schema_, schema.NamedSchema):
            return schema_.fullname if six.PY2 else schema_.fullname.lstrip('.')
        return schema_.type

    def _get_record_schema_if_available(self, data_obj):
        if hasattr(type(data_obj), 'RECORD_SCHEMA'):
            return type(data_obj).RECORD_SCHEMA
        return None

    def _generic_to_json(self, data_obj, writers_schema, was_within_array=False):
        if self.use_logical_types and writers_schema.props.get('logicalType'):
            lt = self.logical_types.get(writers_schema.props.get('logicalType'))  # type: logical.LogicalTypeProcessor
            if lt.can_convert(writers_schema):
                if lt.validate(writers_schema, data_obj):
                    data_obj = lt.convert(writers_schema, data_obj)
                else:
                    raise schema.AvroException(
                        'Wrong object for %s logical type' % writers_schema.props.get('logicalType'))

        if writers_schema.type in _PRIMITIVE_TYPES:
            result = self._primitive_to_json(data_obj, writers_schema)
        elif writers_schema.type == 'fixed':
            result = self._fixed_to_json(data_obj, writers_schema)
        elif writers_schema.type == 'enum':
            result = self._enum_to_json(data_obj, writers_schema)
        elif writers_schema.type == 'array':
            result = self._array_to_json(data_obj, writers_schema)
        elif writers_schema.type == 'map':
            result = self._map_to_json(data_obj, writers_schema)
        elif writers_schema.type in ['record', 'error', 'request']:
            result = self._record_to_json(data_obj, writers_schema)
        elif writers_schema.type in ['union', 'error_union']:
            result = self._union_to_json(data_obj, writers_schema, was_within_array)
        else:
            raise schema.AvroException('Invalid schema type: %s' % writers_schema.type)

        return result

    def _primitive_to_json(self, data_obj, writers_schema):
        return data_obj

    def _fixed_to_json(self, data_obj, writers_schema):
        return data_obj

    def _enum_to_json(self, data_obj, writers_schema):
        return data_obj

    def _array_to_json(self, data_obj, writers_schema):
        return [self._generic_to_json(x, writers_schema.items, True) for x in data_obj]

    def _map_to_json(self, data_obj, writers_schema):
        return {name: self._generic_to_json(x, writers_schema.values) for name, x in six.iteritems(data_obj)}

    def _record_to_json(self, data_obj, writers_schema):
        result = collections.OrderedDict()

        for field in writers_schema.fields:
            result[field.name] = self._generic_to_json(
                data_obj.get(field.name,
                             self.from_json_object(field.default, field.type) if field.has_default else None),
                field.type)
        return result
    
    def _is_unambiguous_union(self, writers_schema) -> bool:
        if any(isinstance(candidate_schema, schema.EnumSchema) for candidate_schema in writers_schema.schemas):
            if any(candidate_schema.type == 'string' for candidate_schema in writers_schema.schemas):
                # Enum and string conflict, so this case is ambiguous.
                return False

        advanced_count = 0
        for candidate_schema in writers_schema.schemas:
            if not isinstance(candidate_schema, schema.PrimitiveSchema):
                advanced_count += 1
        if advanced_count <= 1:
            return True
        return False

    def _union_to_json(self, data_obj, writers_schema, was_within_array=False):
        index_of_schema = -1
        data_schema = self._get_record_schema_if_available(data_obj)
        for i, candidate_schema in enumerate(writers_schema.schemas):
            # Check for exact matches first.
            if data_schema and candidate_schema.namespace == data_schema.namespace \
                and candidate_schema.name == data_schema.name:
                index_of_schema = i
                break

            # Fallback to schema guessing based on validation.
            if self.validate(candidate_schema, data_obj):
                index_of_schema = i
                if candidate_schema.type == 'boolean':
                    break
        if index_of_schema < 0:
            raise io.AvroTypeException(writers_schema, data_obj)
        candidate_schema = writers_schema.schemas[index_of_schema]
        if candidate_schema.type == 'null':
            return None
        
        output_obj = self._generic_to_json(data_obj, candidate_schema)
        if not self.fastavro and not was_within_array and self._is_unambiguous_union(writers_schema):
            # If the union is unambiguous, we can avoid wrapping it in
            # an extra layer of tuples or dicts. Fastavro doesn't like this though.
            # Arrays with unions inside must specify the type.
            return output_obj
        if self.fastavro:
            # Fastavro likes tuples instead of dicts for union types.
            return (self._fullname(candidate_schema), output_obj)
        return {self._fullname(candidate_schema): output_obj}

    def _generic_from_json(self, json_obj, writers_schema, readers_schema):
        if (writers_schema.type not in ['union', 'error_union']
            and readers_schema.type in ['union', 'error_union']):
            for s in readers_schema.schemas:
                if io.DatumReader.match_schemas(writers_schema, s):
                    return self._generic_from_json(json_obj, writers_schema, s)
            raise io.SchemaResolutionException('Schemas do not match', writers_schema, readers_schema)

        result = None
        if writers_schema.type == 'null':
            result = None
        elif writers_schema.type in _PRIMITIVE_TYPES:
            result = self._primitive_from_json(json_obj, writers_schema, readers_schema)
        elif writers_schema.type == 'fixed':
            result = self._fixed_from_json(json_obj, writers_schema, readers_schema)
        elif writers_schema.type == 'enum':
            result = self._enum_from_json(json_obj, writers_schema, readers_schema)
        elif writers_schema.type == 'array':
            result = self._array_from_json(json_obj, writers_schema, readers_schema)
        elif writers_schema.type == 'map':
            result = self._map_from_json(json_obj, writers_schema, readers_schema)
        elif writers_schema.type in ('union', 'error_union'):
            result = self._union_from_json(json_obj, writers_schema, readers_schema)
        elif writers_schema.type in ('record', 'error', 'request'):
            result = self._record_from_json(json_obj, writers_schema, readers_schema)

        result = self._logical_type_from_json(writers_schema, readers_schema, result)
        return result

    def _logical_type_from_json(self, writers_schema, readers_schema, result):
        if self.use_logical_types and readers_schema.props.get('logicalType'):
            lt = self.logical_types.get(readers_schema.props.get('logicalType'))  # type: logical.LogicalTypeProcessor
            if lt and lt.does_match(writers_schema, readers_schema):
                result = lt.convert_back(writers_schema, readers_schema, result)
        return result

    def _primitive_from_json(self, json_obj, writers_schema, readers_schema):
        return json_obj

    def _fixed_from_json(self, json_obj, writers_schema, readers_schema):
        return json_obj

    def _enum_from_json(self, json_obj, writers_schema, readers_schema):
        return json_obj

    def _array_from_json(self, json_obj, writers_schema, readers_schema):
        return [self._generic_from_json(x, writers_schema.items, readers_schema.items)
                for x in json_obj]

    def _map_from_json(self, json_obj, writers_schema, readers_schema):
        return {name: self._generic_from_json(value, writers_schema.values, readers_schema.values)
                for name, value in six.iteritems(json_obj)}

    def _union_from_json(self, json_obj, writers_schema, readers_schema):
        if json_obj is None:
            return None
        value_type = None
        value = None
        if not self.fastavro and isinstance(json_obj, dict):
            items = list(six.iteritems(json_obj))
            if not items:
                return None
            value_type = items[0][0]
            value = items[0][1]
        if self.fastavro and (isinstance(json_obj, list) or isinstance(json_obj, tuple)):
            if len(json_obj) == 2:
                value_type = json_obj[0]
                value = json_obj[1]

        if value_type is not None:
            for s in writers_schema.schemas:
                name = self._fullname(s)
                if name == value_type:
                    return self._generic_from_json(value, s, readers_schema)

        for s in writers_schema.schemas:
            if self.validate(s, json_obj, skip_logical_types=True):
                return self._generic_from_json(json_obj, s, readers_schema)
        raise schema.AvroException('Datum union type not in schema: %s', value_type)
    
    def _make_type(self, tp, record):
        if issubclass(tp, DictWrapper):
            return tp.construct(record)
        return tp(record)

    def _instantiate_record(self, decoded_record, writers_schema, readers_schema):
        # First try the fullname, which includes namespaces.
        readers_name = self._fullname(readers_schema)
        if readers_name in self.schema_types:
            return self._make_type(self.schema_types[readers_name], decoded_record)
        # Fallback to the bare name, without namespace.
        readers_name = readers_schema.name
        if readers_name in self.schema_types:
            return self._make_type(self.schema_types[readers_name], decoded_record)
        return decoded_record

    def _record_from_json(self, json_obj, writers_schema, readers_schema):
        writer_fields = writers_schema.fields_dict if six.PY2 else writers_schema.field_map

        result = {}
        for field in readers_schema.fields:
            writers_field = writer_fields.get(field.name)
            if writers_field is None:
                field_value = self._generic_from_json(field.default, field.type, field.type) \
                    if field.has_default else None
            else:
                if field.name in json_obj:
                    field_value = self._generic_from_json(json_obj[field.name], writers_field.type, field.type)
                else:
                    field_value = self._generic_from_json(writers_field.default,
                                                          writers_field.type, field.type) \
                        if writers_field.has_default else None
            result[field.name] = field_value
        return self._instantiate_record(result, writers_schema, readers_schema)



"""
A setuptools-based setup module.

See:
https://github.com/hsheth2/avro_gen
"""

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
if path.exists(path.join(here, 'README.md')):
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
else:
    long_description = ''

setup(
    name='avro-gen3',
    version='0.4.2',

    description='Avro record class and specific record reader generator',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/hsheth2/avro_gen',
    author='Harshal Sheth',
    author_email='hsheth2@gmail.com',
    license='License :: OSI Approved :: Apache Software License',

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],

    keywords='avro class generator',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    package_data={
        'avrogen': ['py.typed'],
    },
    install_requires=["avro >= 1.8.0 ; python_version<'3.0'",
                      "avro_python3 >= 1.8.0 ; python_version>'3.0'",
                      'six', 'tzlocal', 'pytz'],
)



from .schema import generate_schema,  write_schema_files
from .protocol import generate_protocol, write_protocol_files
__all__ = ['generate_schema', 'generate_protocol', 'write_schema_files', 'write_protocol_files']


